/* 
 * Desc:
 *
 * Created by Magnus Berglund
 * 
 * Date: 
 */

#define VERSION "1.2.2"

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include "truesound.h"

#define error(args...) ({ fprintf(stderr, args); exit(1); })

#define MIN(A, B) ((A)<(B)?(A):(B))
#define MAX(A, B) ((A)>(B)?(A):(B))

static int device_fileno=0;
static unsigned long buffsize;
static int bytes_in_buff=0;

static char *currentfile;
static int fileindex;
static int numberoffiles = 0;

static char *progname;


/* command line arguments */
static int repeat=0;
static long skip_size=0;
static int byteprompt=0;
static int spaceskip=0;
static int defaultlayer=2;
static int skip=5;
static int prompt_type=0;
static int device=0;


extern inline int isemptyfilename(const char *name)
{
    int i;
    int len = strlen(name);
    for (i=0; i<len; i++) 
	if (name[i]!=' ' && name[i]!='\t') return 0;
    return 1;
}

extern inline int isint(const char *var)
{
    int i=0;
    while(var[i]!=0) 
	if (!isdigit(var[i++])) return 0;
    return 1;
}

void play_pause(void)
{
    int ret;
    ret = ioctl(device_fileno, TRUESOUND_PAUSE);
    if (-1==ret) { perror("ioctl(TRUESOUND_PAUSE)"); exit(1); }
}

void play_cont(void)
{
    int ret;
    ret = ioctl(device_fileno, TRUESOUND_CONT);
    if (-1==ret) { perror("ioctl(TRUESOUND_CONT)"); exit(1); }
}

void decrement_volume(void)
{
    unsigned char volume;
    int ret;
    ret = ioctl(device_fileno, TRUESOUND_GETVOL, &volume);
    if (-1==ret) { perror("ioctl(TRUESOUND_GETVOL)"); exit(1); }
    volume = MAX(0, volume-5);
    ret = ioctl(device_fileno, TRUESOUND_SETVOL, &volume);
    if (-1==ret) { perror("ioctl(TRUESOUND_SETVOL)"); exit(1); }
}

void increment_volume(void)
{
    unsigned char volume;
    int ret;
    ret = ioctl(device_fileno, TRUESOUND_GETVOL, &volume);
    if (-1==ret) { perror("ioctl(TRUESOUND_GETVOL)"); exit(1); }
    volume = MIN(100, volume+5);
    ret = ioctl(device_fileno, TRUESOUND_SETVOL, &volume);
    if (-1==ret) { perror("ioctl(TRUESOUND_SETVOL)"); exit(1); }
}

void flush_buffers(void)
{
    if (device_fileno) {
	int ret;
	ret = ioctl(device_fileno, TRUESOUND_FLUSH_BUFFERS);
	if (-1==ret) { perror("ioctl(TRUESOUND_FLUSH_BUFFERS)"); exit(1); }
	bytes_in_buff=0;
    }
}

void quit(int sig)
{
    flush_buffers();
    fputs("\n", stdout);
    exit(1);
}



#define QUIT 1
#define NEXT 2
#define PREV 3
#define FORWARD 4
#define BACKWARD 5
int handle_keyb_command(void) {
    char c;
    int ret;
    ret = read(STDIN_FILENO, &c, 1);
    if (-1==ret) {
	perror("read(stdin)");
	exit(0);
    }
    if (1==ret) {
	c=tolower(c);
	switch(c) {
	case '-' : decrement_volume(); break;
	case '+' : increment_volume(); break;
	case 'p' : play_pause(); break;
	case 'c' : play_cont(); break;
	case 'f' : return FORWARD;
	case 'b' : return BACKWARD;
	case 'q' : return QUIT;
	case 'n' : return NEXT;
	case 'r' : return PREV;
	default: break;
	}
    }
    return 0;
}


#define PROGRESS_SIZE 20
void update_prompt(unsigned long total_bytes_written, 
		   unsigned long file_size, int force_update) 
{
    unsigned char vol;
    int paused;
    unsigned long done;
    static unsigned char old_vol=0;
    static int old_paused=2;
    static unsigned long old_done=100;
    static unsigned long old_total_bytes_written=100;

    static char progress[PROGRESS_SIZE+1];
    int ret;

    ret = ioctl(device_fileno, TRUESOUND_GETVOL, &vol);
    if (-1==ret) {
	perror("ioctl(TRUESOUND_GETVOL)");
	exit(1);
    }

    ret = ioctl(device_fileno, TRUESOUND_GET_PAUSED, &paused);
    if (-1==ret) {
	perror("ioctl(TRUESOUND_GET_PAUSED)");
	exit(1);
    }
    
    done = (total_bytes_written * (PROGRESS_SIZE+1)) / file_size;
    done = MIN(PROGRESS_SIZE,done);

    if (!byteprompt) {
	if (!force_update 
	    && vol==old_vol 
	    && paused==old_paused 
	    && done==old_done) return;
    }
    else { 
	if (!force_update 
	    && vol==old_vol 
	    && paused==old_paused 
	    && total_bytes_written==old_total_bytes_written)
	    return;
    }
    
    memset(progress, '-', PROGRESS_SIZE);
    memset(progress, '=', done);
    progress[PROGRESS_SIZE]='\0';

    if (!byteprompt) {
	if (!paused)
	    printf("\r[%d/%d] [%s] %d%% "
		   "[+/-/Pause/Fwd/Bck/Next/pRevious/Quit] ", 
		   fileindex+1, numberoffiles, progress, (int)vol);
	else
	    printf("\r[%d/%d] [%s] %d%% "
		   "[+/-/Cont/Fwd/Bck/Next/pRevious/Quit] ", 
		   fileindex+1, numberoffiles, progress, (int)vol);
    }
    else {
	if (!paused)
	    printf("\r[%d/%d] [%lu/%lu] %d%% "
		   "[+/-/Pause/Fwd/Bck/Next/pRevious/Quit] ", 
		   fileindex+1, numberoffiles, total_bytes_written, 
		   file_size, (int)vol);
	else
	    printf("\r[%d/%d] [%lu/%lu] %d%% "
		   "[+/-/Cont/Fwd/Bck/Next/pRevious/Quit] ", 
		   fileindex+1, numberoffiles, total_bytes_written, 
		   file_size, (int)vol);
    }

    fflush(stdout);

    old_vol=vol;
    old_paused=paused;
    old_done=done;
    old_total_bytes_written=total_bytes_written;
}



int playfile(int in_fileno, unsigned long file_size) 
{

#define BLOCKSIZE 4096
    ssize_t read_ret;
    ssize_t write_ret;
    ssize_t totalcount = 0;
    char buffer[BLOCKSIZE];
    ssize_t write_count;
    int sel_ret;
    fd_set rds, wds;
    struct timeval timeout;

    do {
	read_ret = read(in_fileno, buffer, BLOCKSIZE);

	if (read_ret>0) {
	    write_count = 0;

	    while (write_count<read_ret) {
		
		FD_ZERO(&wds);
		FD_SET(device_fileno, &wds);
		FD_ZERO(&rds);
		FD_SET(STDIN_FILENO, &rds);
		timeout.tv_sec = 2;
		timeout.tv_usec = 0;
		
		sel_ret = select(device_fileno+1, &rds, &wds, NULL, &timeout);
		if (0==sel_ret) {
		    update_prompt((unsigned long)totalcount, 
				  file_size, 0);
		}
		else if (-1==sel_ret) {
		    if (EINTR!=errno) {
			perror("select");
			return QUIT;
		    }
		}
		else {
		    if (FD_ISSET(device_fileno, &wds)) {
			write_ret = write(device_fileno, &buffer[write_count], 
					  read_ret-write_count);
			if (-1==write_ret) {
			    if (EINTR!=errno) {
				perror("write(device)");
				return QUIT;
			    }
			}
			else {
			    bytes_in_buff = MIN(buffsize, 
						bytes_in_buff+write_ret);
			    totalcount+=write_ret;
			    write_count+=write_ret;
			    update_prompt((unsigned long)totalcount,
					  file_size,0);
			}
		    }
		    if (FD_ISSET(STDIN_FILENO, &rds)) {
			int ret;
			off_t lret;
			ret = handle_keyb_command();
			if (FORWARD==ret) {
			    lret=lseek(in_fileno, skip_size-bytes_in_buff,
				       SEEK_CUR);
			    flush_buffers(); 
			    if (-1!=lret) { 
				totalcount=lret;
				write_count=read_ret;
			    }
			}
			else if (BACKWARD==ret) {
			    lret=lseek(in_fileno, -skip_size-bytes_in_buff,
				       SEEK_CUR);
			    flush_buffers(); 
			    if (-1!=lret) { 
				totalcount=lret;
				write_count=read_ret;
			    }
			}
			else if (ret) return ret;
			update_prompt((unsigned long)totalcount, 
				      file_size, 0);
		    }
		}
	    }
	}
	else if (-1==read_ret) {
	    perror(currentfile);
	    return 0;
	}
    } while(read_ret!=0);
    
    return 0;
}



char **add_filenames(const char *file, char **names, int *number)
{
    static char tmpbuff[1024];
    FILE *tmpfile;
    char *newname;
    int newcount;
    int lines = 0;

    tmpfile = fopen(file, "r");
    if (!tmpfile) {
	perror(file);
	exit(2);
    }
    while (fgets(tmpbuff, 1024, tmpfile)) 
	lines++;
    names = (char **)realloc(names, sizeof(char *) * (*number+lines));
    if (!names) error("out of memory");

    rewind(tmpfile);

    for (newcount=*number;;) {
	newname = fgets(tmpbuff, 1024, tmpfile);
	if (!newname) break;
	if ('\n'==newname[strlen(newname)-1])
	    newname[strlen(newname)-1]='\0';
	if (isemptyfilename(newname)) continue;
	if (spaceskip) {
	    char *space = strchr(newname, ' ');
	    if (space) *space='\0';
	}
	names[newcount++] = strdup(newname);
    }
    *number = newcount;
    fclose(tmpfile);
    return names;
}

int check_file(char *filename) {
    struct stat sbuf;
    int ret;
    ret = lstat(filename, &sbuf);
    if (ret!=0) { 
	perror(filename);
	return 1;
    }
    if (access(filename, R_OK)!=0) {
	perror(filename);
	return 1;
    }		
    if (!S_ISREG(sbuf.st_mode) && !S_ISLNK(sbuf.st_mode)) {
	fprintf(stderr, "%s: is a not a regular file\n", filename);
	return 1;
    }
    return 0;
}


void check_that_files_readable(char **files, int *number)
{
    int filecnt = 0;
    int i, ret;

    if (*number>200) 
	printf("-- checking that the files exists and are readable --\n");

    for (i=0; i<*number; i++) {
	ret = check_file(files[i]);
	if (0!=ret) continue;
	files[filecnt]=files[i];
	filecnt++;
    }
    *number=filecnt;
}



extern inline void swap(char **a, char **b) {
    char *tmp;
    tmp = *a;
    *a = *b;
    *b = tmp;
}


void randomize_files(char **filenames, int numberoffiles) 
{
    int i;
    int rand;
    struct timeval tval;

    if (numberoffiles>200) 
	printf("-- randomizing files --\n");

    gettimeofday(&tval, NULL);
    srandom((int)tval.tv_usec);
    for (i=0; i<numberoffiles; i++) {
	rand = (int)random() % numberoffiles;
	swap(&filenames[i], &filenames[rand]);
    }
}


int play_files(char **filenames, int numberoffiles)
{
    static const char *devname[3][3] = { 
	{"/dev/tsd0-mp1", "/dev/tsd0-mp2", "/dev/tsd0-mp3" },
	{"/dev/tsd1-mp1", "/dev/tsd1-mp2", "/dev/tsd1-mp3" },
	{"/dev/tsd2-mp1", "/dev/tsd2-mp2", "/dev/tsd2-mp3" },
    };
	
    struct stat b_fstat;
    int layer, in_fd, len, i, ret;

    fileindex = 0;

    if (3==prompt_type) {
	for (i=0; i<numberoffiles; i++) {
	    char *filename2 = strrchr(filenames[i], '/');
	    printf("%2d - %s\n", i+1, filename2 ? filename2+1 : filenames[i]);
	}
    }
    
    for (;;) {

	if (repeat && fileindex>=numberoffiles) fileindex=0;
	if (fileindex<0) 
	    if (repeat) 
		fileindex=numberoffiles-1;
	    else
		fileindex=0;
	if (fileindex>=numberoffiles) break;

	currentfile = filenames[fileindex];

	if (check_file(currentfile)) {
	    fileindex++;
	    continue;
	}

	len = strlen(currentfile);

	in_fd = open(currentfile, O_RDONLY); 
	if (-1==in_fd) {
	    perror(currentfile);
	    fileindex++;
	    continue;
	}

	fstat(in_fd, &b_fstat);

        if(prompt_type==0) {
	    printf("%s:\n", currentfile);
	}
	else if(prompt_type==1) {
	    char *filename2 = strrchr(currentfile, '/');
	    printf("%s:\n", filename2 ? filename2+1 : currentfile);
	}

	if     (len>3 && strcasecmp(&currentfile[len-4],".mp1")==0)   layer=0;
	else if(len>3 && strcasecmp(&currentfile[len-4],".m1a")==0)   layer=0;
	else if(len>4 && strcasecmp(&currentfile[len-5],".mpg1")==0)  layer=0;
	else if(len>5 && strcasecmp(&currentfile[len-6],".mpeg1")==0) layer=0;
	else if(len>3 && strcasecmp(&currentfile[len-4],".mp2")==0)   layer=1;
	else if(len>3 && strcasecmp(&currentfile[len-4],".m2a")==0)   layer=1;
	else if(len>4 && strcasecmp(&currentfile[len-5],".mpg2")==0)  layer=1;
	else if(len>5 && strcasecmp(&currentfile[len-6],".mpeg2")==0) layer=1;
	else if(len>3 && strcasecmp(&currentfile[len-4],".mp3")==0)   layer=2;
	else if(len>3 && strcasecmp(&currentfile[len-4],".m3a")==0)   layer=2;
	else if(len>4 && strcasecmp(&currentfile[len-5],".mpg3")==0)  layer=2;
	else if(len>5 && strcasecmp(&currentfile[len-6],".mpeg3")==0) layer=2;
	else { 
	    layer=defaultlayer;
	    printf("%s: assuming that this is layer %d stream...\n", 
		   currentfile, layer+1);
	}
	
	device_fileno = open(devname[device][layer], O_WRONLY);
	if (-1==device_fileno) {
	    perror(devname[device][layer]);
	    close(in_fd);
	    exit(2);
	}

	bytes_in_buff=0;
	ret = ioctl(device_fileno, TRUESOUND_GET_BUFFSIZE, &buffsize);
	if (-1==ret) { perror("ioctl(TRUESOUND_GET_BUFFSIZE)"); exit(1); }
	skip_size = (b_fstat.st_size*skip) / 100;

	update_prompt(0L, (unsigned long)b_fstat.st_size, 1);

	ret = playfile(in_fd, (unsigned long)b_fstat.st_size);
	switch(ret) {
	case 0:
	    fileindex++;
	    break;
	case NEXT:
	    flush_buffers();
	    fileindex++;
	    break;
	case PREV:
	    flush_buffers();
	    fileindex--;
	    break;
	case QUIT:
	    quit(0);
	}
	close(in_fd);
	close(device_fileno);
	device_fileno=0;
	if (0==prompt_type || 1==prompt_type)
	    fputs("\n", stdout);
    }

    if (2==prompt_type || 3==prompt_type)
	fputs("\n", stdout);

    return 0;
    
}


static struct termios termorig;
void restore_term(void)
{
    tcsetattr(STDIN_FILENO, TCSANOW, &termorig);
}




void usage(const char *prgname) {
    fprintf(stderr, 
	    "Usage: %s [OPTION...] files...\n"
	    "Play files on the truesound decoder device\n"
	    "  -d, --device         choose card (0,1,2) default 0\n"
	    "  -f, --file=FILE      read filenames from 'FILE'\n"
	    "  -s, --spaceskip      skip averything after a ' ' in 'FILE'\n"
	    "  -r, --repeat         repeat song(s)\n"
	    "  -b, --byteprompt     view bytes written instead of progress bar\n"
	    "  -n, --nofilecheck    Do not check that files are readable before playing\n" 
	    "      --random         randomize songs\n"
	    "      --skip=SKIP      skip 'SKIP' percent of the song then fast forwarding\n"
	    "  -p, --prompt=TYPE,   0 - print complete filename before echo song\n"
	    "                       1 - print short filename before echo song\n"
	    "                       2 - do not print any filename(s)\n"
	    "                       3 - print filenames once\n"
	    "  -1,                  assume layer 1 for files with unknown extention\n"
	    "  -2,                  assume layer 2 for files with unknown extention\n"
	    "  -3,                  assume layer 3 for files with unknown extention\n"
	    "  -h, --help           print this help\n"
	    "      --version        print version\n",
	    prgname);
    
}





int main(int argc, char **argv)
{
    int opt, i, ret;
    struct termios term;
    char *listoffiles[100];
    int listoffilescount=0;
    int random = 0;
    int nofilecheck = 0;

    char **filenames = NULL;

    static struct option long_options[]=
    {
	{"device",  1, 0, 'd'},
	{"file", 1, 0, 'f'},
	{"spaceskip", 0, 0, 's'},
	{"random", 0, 0, 3},
	{"repeat",  0, 0, 'r'},
	{"nofilecheck", 0, 0, 'n'},
	{"byteprompt", 0, 0, 'b'},
	{"skip", 1, 0, 2},
	{"prompt", 1, 0, 'p'},
	{"help",    0, 0, 'h'},
	{"version", 0 ,0, 1},
	{0, 0, 0, 0}
    };
    
    progname=argv[0];
    listoffiles[0] = NULL;

    while ( (opt=getopt_long(argc, argv, "d:f:rnbs123hp:", long_options, 0))!=-1)
	{
	    switch(opt)
		{
		case 'd':
		    if (!isint(optarg)) 
			error("%s: invalid device: %s\n", 
			      progname, optarg);
		    device = atoi(optarg);
		    if (device<0 || device>2)
			error("%s: device option must be 0,1 or 2\n",
			      progname);
		    break;
		case 'f':
		    listoffiles[listoffilescount++] = optarg;
		    break;
		case 's':
		    spaceskip=1;
		    break;
		case 3:
		    random=1;
		    break;
		case 'r': 
		    repeat=1;
		    break;
		case 'n':
		    nofilecheck=1;
		    break;
		case 'b':
		    byteprompt=1;
		    break;
		case 2:
		    if (!isint(optarg))
			error("%s: invalid skip: %s\n", 
			      progname, optarg);
		    skip = atoi(optarg);
		    if (skip<0 || skip>100)
			error("%s: skip must be in range 0-100\n", progname);
		    break;

		case 'p': 
		    if (!isint(optarg))
			error("%s: invalid prompttype: %s\n", 
			      progname, optarg);
		    prompt_type = atoi(optarg);
		    if (prompt_type<0 || prompt_type>3) {
			error("%s: prompttype must be in range 0-3\n", 
			      progname);
		    }
		    break;
			
		case '1':
		    defaultlayer=0;
		    break;
		case '2':
		    defaultlayer=1;
		    break;
		case '3':
		    defaultlayer=2;
		    break;
		case 'h':
		    usage(progname);
		    exit(0);
		case 1:
		    fprintf(stderr, "%s version %s, compiledate %s\n", 
			    progname, VERSION, __DATE__);
		    exit(0);
		default:
		    error("Try `%s --help` for more information\n", 
			  progname);
		}
	    
	}

    for (i=0; i<listoffilescount; i++) 
	filenames = add_filenames(listoffiles[i], filenames, &numberoffiles);

    if (argc-optind) {
	filenames = realloc(filenames, 
			    sizeof(char *)*(numberoffiles+argc-optind));
	if (NULL==filenames) error("ouf of memory\n");
	for (i=0; i<(argc-optind); i++) {
	    filenames[numberoffiles++] = argv[optind+i];
	}
    }

    if (0==nofilecheck) 
	check_that_files_readable(filenames, &numberoffiles);

    if (0==numberoffiles) 
	error("Failed at: no files to play.\n");

    if (random)
	randomize_files(filenames, numberoffiles);
    
    ret = tcgetattr(STDIN_FILENO, &termorig);
    if (-1==ret) {
	if (EINVAL==errno) error("%s: stdin is not a tty\n", progname);
	else {
	    perror("tcgetattr");
	    exit(1);
	}
    }
    atexit(restore_term);
    memcpy(&term, &termorig, sizeof(struct termios));
    term.c_lflag &= ~(ICANON | ECHO);
    ret = tcsetattr(STDIN_FILENO, TCSANOW, &term);
    if (-1==ret) {
	perror("tcgetattr");
	exit(1);
    }

    signal(SIGINT, quit);
    signal(SIGQUIT, quit);
    signal(SIGTERM, quit);
    
    play_files(filenames, numberoffiles);

    return 0;
}
